package ru.dbotthepony.kstarbound.io

import org.lwjgl.stb.STBImage
import org.lwjgl.system.MemoryUtil.memAddress
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.nint.Vector2i
import java.lang.ref.Cleaner
import java.nio.ByteBuffer

private val cleaner = Cleaner.create { Thread(it, "STB Image Cleaner") }

class ImageData(val data: ByteBuffer, val width: Int, val height: Int, val amountOfChannels: Int) {
	init {
		val address = memAddress(data)
		cleaner.register(data) { STBImage.nstbi_image_free(address) }

		check(width >= 0) { "Invalid width $width" }
		check(height >= 0) { "Invalid height $height" }
		check(amountOfChannels in 1 .. 4) { "Unknown number of channels $amountOfChannels" }
	}

	operator fun get(x: Int, y: Int): Int {
		require(x in 0 until width) { "X out of bounds: $x" }
		require(y in 0 until height) { "Y out of bounds: $y" }

		val offset = y * width * amountOfChannels + x * amountOfChannels

		when (amountOfChannels) {
			4 -> return data[offset].toInt() or
				data[offset + 1].toInt().shl(8) or
				data[offset + 2].toInt().shl(16) or
				data[offset + 3].toInt().shl(24)

			3 -> return data[offset].toInt() or
				data[offset + 1].toInt().shl(8) or
				data[offset + 2].toInt().shl(16)

			2 -> return data[offset].toInt() or
				data[offset + 1].toInt().shl(8)

			1 -> return data[offset].toInt()

			else -> throw IllegalStateException()
		}
	}

	fun worldSpaces(pos: Vector2d, spaceScan: Double, flip: Boolean): List<Vector2i> {
		when (amountOfChannels) {
			3 -> TODO()
			4 -> TODO()
			else -> throw IllegalStateException("Can not check world space taken by image with $amountOfChannels color channels")
		}
	}
}
